@@ -1,12 +1,28 @@
 import math
 import types
 from datetime import datetime, timedelta
-from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    List,
+    Optional,
+    Tuple,
+    Type,
+    Union,
+)
 
-import altair as alt
+if TYPE_CHECKING:
+    import altair as alt
+
+import numpy as np
 import pandas as pd
 import pytz
 from pandas.core.groupby import DataFrameGroupBy
+from pandas.tseries.frequencies import to_offset
+from sktime.forecasting.base import BaseForecaster
+from sktime.forecasting.naive import NaiveForecaster
 from sqlalchemy import (
     Column,
     DateTime,
@@ -34,7 +50,6 @@ from timely_beliefs.sensors import utils as sensor_utils
 from timely_beliefs.sensors.classes import DBSensor, Sensor, SensorDBMixin
 from timely_beliefs.sources import utils as source_utils
 from timely_beliefs.sources.classes import BeliefSource, DBBeliefSource
-from timely_beliefs.visualization import utils as visualization_utils
 
 METADATA = ["sensor", "event_resolution"]
 DatetimeLike = Union[datetime, str, pd.Timestamp]
@@ -100,7 +115,7 @@ class TimedBelief(object):
         elif cp is not None:
             self.cumulative_probability = cp
         elif sigma is not None:
-            self.cumulative_probability = 1 / 2 + (math.erf(sigma / 2 ** 0.5)) / 2
+            self.cumulative_probability = 1 / 2 + (math.erf(sigma / 2**0.5)) / 2
         else:
             self.cumulative_probability = 0.5
 
@@ -245,7 +260,7 @@ class TimedBeliefDBMixin(TimedBelief):
     def add_to_session(
         cls,
         session: Session,
-        beliefs_data_frame: "BeliefsDataFrame",  # todo: allow passing List[TimedBelief], too
+        beliefs_data_frame: "BeliefsDataFrame",
         expunge_session: bool = False,
         allow_overwrite: bool = False,
         bulk_save_objects: bool = False,
@@ -291,22 +306,6 @@ class TimedBeliefDBMixin(TimedBelief):
         if commit_transaction:
             session.commit()
 
-    @classmethod
-    def expunge_from_session(
-        cls,
-        session: Session,
-        beliefs_data_frame: "BeliefsDataFrame",  # todo: allow passing List[TimedBelief], too
-    ):
-        # Belief timing is stored as the belief horizon rather than as the belief time
-        belief_records = (
-            beliefs_data_frame.convert_index_from_belief_time_to_horizon()
-                .reset_index()
-                .to_dict("records")
-        )
-        beliefs = [cls(sensor=beliefs_data_frame.sensor, **d) for d in belief_records]
-        for belief in beliefs:
-            session.expunge(belief)
-
     @classmethod
     @tb_utils.append_doc_of("TimedBeliefDBMixin.search_session")
     def query(cls, *args, **kwargs):
@@ -411,19 +410,19 @@ class TimedBeliefDBMixin(TimedBelief):
         )
 
         # Check for timezone-aware datetime input
-        if event_starts_after is not None:
+        if not pd.isnull(event_starts_after):
             event_starts_after = tb_utils.parse_datetime_like(
                 event_starts_after, "event_not_before"
             )
-        if event_ends_before is not None:
+        if not pd.isnull(event_ends_before):
             event_ends_before = tb_utils.parse_datetime_like(
                 event_ends_before, "event_before"
             )
-        if beliefs_after is not None:
+        if not pd.isnull(beliefs_after):
             beliefs_after = tb_utils.parse_datetime_like(
                 beliefs_after, "belief_not_before"
             )
-        if beliefs_before is not None:
+        if not pd.isnull(beliefs_before):
             beliefs_before = tb_utils.parse_datetime_like(
                 beliefs_before, "belief_before"
             )
@@ -458,31 +457,35 @@ class TimedBeliefDBMixin(TimedBelief):
         q = session.query(cls).filter(cls.sensor_id == sensor.id)
 
         # Apply event time filter
-        if event_starts_after is not None:
+        if not pd.isnull(event_starts_after):
             q = q.filter(cls.event_start >= event_starts_after)
-        if event_ends_before is not None:
+        if not pd.isnull(event_ends_before):
             q = q.filter(cls.event_start + sensor.event_resolution <= event_ends_before)
 
         # Apply rough belief time filter
-        if beliefs_after is not None:
+        if not pd.isnull(beliefs_after) and belief_utils.extreme_timedeltas_not_equal(
+            knowledge_horizon_min, timedelta.min
+        ):
             q = q.filter(
                 cls.event_start
                 >= beliefs_after + cls.belief_horizon + knowledge_horizon_min
             )
-        if beliefs_before is not None:
+        if not pd.isnull(beliefs_before) and belief_utils.extreme_timedeltas_not_equal(
+            knowledge_horizon_max, timedelta.max
+        ):
             q = q.filter(
                 cls.event_start
                 <= beliefs_before + cls.belief_horizon + knowledge_horizon_max
             )
 
         # Apply belief horizon filter
-        if horizons_at_least is not None:
+        if not pd.isnull(horizons_at_least):
             q = q.filter(cls.belief_horizon >= horizons_at_least)
-        if horizons_at_most is not None:
+        if not pd.isnull(horizons_at_most):
             q = q.filter(cls.belief_horizon <= horizons_at_most)
 
         # Apply source filter
-        if source is []:
+        if source == []:
             return BeliefsDataFrame(sensor=sensor, beliefs=[])
         elif source is not None:
             sources: list = [source] if not isinstance(source, list) else source
@@ -702,6 +705,25 @@ class BeliefsDataFrame(pd.DataFrame):
     def __finalize__(self, other, method=None, **kwargs):
         """Propagate metadata from other to self."""
         # merge operation: using metadata of the left object
+
+        # Check if sources have unique names
+        if hasattr(other, "objs"):
+            sources = []
+            for df in other.objs:
+                if "source" in df.index:
+                    sources.extend(
+                        df.index.get_level_values(level="source")
+                        .unique()
+                        .to_numpy(dtype="object")
+                    )
+            sources = set(sources)
+            source_names = set(source.name for source in sources)
+            if len(source_names) != len(sources):
+                raise ValueError(
+                    "Source names must be unique. Cannot initialise BeliefsDataFrame given the following unique sources:\n%s"
+                    % sources
+                )
+
         if method == "merge":
             for name in self._metadata:
                 object.__setattr__(self, name, getattr(other.left, name, None))
@@ -777,7 +799,7 @@ class BeliefsDataFrame(pd.DataFrame):
             sensor = list(unique_sensors)[0]
 
             # Check for different sources with the same name
-            unique_sources = set(belief.source for belief in beliefs)
+            unique_sources = set(str(belief.source) for belief in beliefs)
             unique_source_string_representations = set(
                 str(source) for source in unique_sources
             )
@@ -1128,6 +1150,10 @@ class BeliefsDataFrame(pd.DataFrame):
         """
         return self._for_each_belief(fnc, True, *args, **kwargs)
 
+    @hybrid_method
+    def make_deterministic(self):
+        return self.for_each_belief(probabilistic_utils.get_expected_belief)
+
     @hybrid_method
     def belief_history(
         self,
@@ -1327,11 +1353,16 @@ class BeliefsDataFrame(pd.DataFrame):
         event_resolution: TimedeltaLike,
         distribution: Optional[str] = None,
         keep_only_most_recent_belief: bool = False,
+        keep_nan_values: bool = False,
     ) -> "BeliefsDataFrame":
         """Aggregate over multiple events (downsample) or split events into multiple sub-events (upsample).
 
+        Drops NaN values by default.
+
         NB If you need to only keep the most recent belief,
         set keep_only_most_recent_belief=True for a significant speed boost.
+
+        :param keep_nan_values: if True, place back resampled NaN values.
         """
 
         if self.empty:
@@ -1345,17 +1376,19 @@ class BeliefsDataFrame(pd.DataFrame):
             "belief_time" if "belief_time" in df.index.names else "belief_horizon"
         )
 
-        # fast track a common case where each event has only one deterministic belief and only the most recent belief is needed
-        print("bbbbbbbbb")
-        print(df.lineage.number_of_beliefs == df.lineage.number_of_events)
-        print(keep_only_most_recent_belief)
-        print( df.lineage.number_of_sources)
+        # fast track a common case where each event has only one deterministic belief by the same source, and:
+        # - only the most recent belief is needed, or
+        # - we are upsampling, or
+        # - all beliefs share a common belief time
         if (
             df.lineage.number_of_beliefs == df.lineage.number_of_events
-            and keep_only_most_recent_belief
+            and (
+                keep_only_most_recent_belief
+                or event_resolution < self.event_resolution
+                or df.lineage.number_of_belief_times == 1
+            )
             and df.lineage.number_of_sources == 1
         ):
-            print("rrrrrrrrrr")
             if event_resolution > self.event_resolution:
                 # downsample
                 column_functions = {
@@ -1364,7 +1397,7 @@ class BeliefsDataFrame(pd.DataFrame):
                     belief_timing_col: "max"
                     if belief_timing_col == "belief_time"
                     else "min",  # keep only most recent belief
-                    "cumulative_probability": "prod",  # assume independent variables
+                    "cumulative_probability": "mean",  # we just have one point on each CDF
                 }
                 df = downsample_beliefs_data_frame(
                     df, event_resolution, column_functions
@@ -1375,13 +1408,55 @@ class BeliefsDataFrame(pd.DataFrame):
                 df = df.reset_index(
                     level=[belief_timing_col, "source", "cumulative_probability"]
                 )
-                new_index = pd.date_range(
+                resample_ratio = pd.to_timedelta(
+                    to_offset(df.event_resolution)
+                ) / pd.Timedelta(event_resolution)
+                if keep_nan_values:
+                    # back up NaN values
+                    unique_event_value_not_in_df = df["event_value"].abs().sum() + 1
+                    df = df.fillna(unique_event_value_not_in_df)
+                new_index = belief_utils.initialize_index(
                     start=df.index[0],
-                    periods=len(df) * (self.event_resolution // event_resolution),
-                    freq=event_resolution,
-                    name="event_start",
+                    end=df.index[-1] + self.event_resolution,
+                    resolution=event_resolution,
                 )
-                df = df.reindex(new_index).fillna(method="pad")
+                # Reindex to introduce NaN values, then forward fill by the number of steps
+                # needed to have the new resolution cover the old resolution.
+                # For example, when resampling from a resolution of 30 to 20 minutes (NB frequency is 1 hour):
+                # event_start               event_value
+                # 2020-03-29 10:00:00+02:00 1000.0
+                # 2020-03-29 11:00:00+02:00 NaN
+                # 2020-03-29 12:00:00+02:00 2000.0
+                # After reindexing
+                # event_start               event_value
+                # 2020-03-29 10:00:00+02:00 1000.0
+                # 2020-03-29 10:20:00+02:00 NaN
+                # 2020-03-29 10:40:00+02:00 NaN
+                # 2020-03-29 11:00:00+02:00 NaN
+                # 2020-03-29 11:20:00+02:00 NaN
+                # 2020-03-29 11:40:00+02:00 NaN
+                # 2020-03-29 12:00:00+02:00 2000.0
+                # 2020-03-29 12:20:00+02:00 NaN
+                # 2020-03-29 12:40:00+02:00 NaN
+                # After filling a limited number of NaN values (ceil(30/20)-1 == 1)
+                # event_start               event_value
+                # 2020-03-29 10:00:00+02:00 1000.0
+                # 2020-03-29 10:20:00+02:00 1000.0
+                # 2020-03-29 10:40:00+02:00 NaN
+                # 2020-03-29 11:00:00+02:00 NaN
+                # 2020-03-29 11:20:00+02:00 NaN
+                # 2020-03-29 11:40:00+02:00 NaN
+                # 2020-03-29 12:00:00+02:00 2000.0
+                # 2020-03-29 12:20:00+02:00 2000.0
+                # 2020-03-29 12:40:00+02:00 NaN
+                df = df.reindex(new_index).fillna(
+                    method="pad",
+                    limit=math.ceil(resample_ratio) - 1 if resample_ratio > 1 else None,
+                )
+                df = df.dropna()
+                if keep_nan_values:
+                    # place back original NaN values
+                    df = df.replace(unique_event_value_not_in_df, np.NaN)
                 df.event_resolution = event_resolution
                 df = df.set_index(
                     [belief_timing_col, "source", "cumulative_probability"], append=True
@@ -1419,6 +1494,140 @@ class BeliefsDataFrame(pd.DataFrame):
 
         return df
 
+    @hybrid_method
+    def form_beliefs(
+        self,
+        belief_time: datetime,
+        source: BeliefSource,
+        event_start: datetime = None,
+        event_time_window: Optional[Tuple[datetime, datetime]] = (
+            None,
+            None,
+        ),
+        forecaster: Optional[BaseForecaster] = None,
+        concatenate: bool = False,
+    ):
+        """Form new beliefs by applying a given forecaster.
+
+        Note that the result may contain NaN values, for example, when the BeliefsDataFrame contains insufficient data.
+
+        :param belief_time:       Time at which the forecasts where made
+                                  (any belief after this time will be inaccessible to the forecaster).
+        :param source:            Source to assign the newly formed beliefs to.
+        :param event_start:       Set this to forecast a single event with the given start time.
+        :param event_time_window: Set this to forecast all events within the given time window.
+        :param forecaster:        Forecasting model. Currently, only sktime models are supported.
+                                  The default forecaster simply repeats the last known value.
+                                  Hint: set a seasonal periodicity to obtain a more reasonable (baseline) forecast.
+                                  For example:
+                                      >>> periodicity = pd.Timedelta("PT1W")
+                                      >>> forecaster = NaiveForecaster(sp=periodicity // df.event_resolution)
+        :param concatenate:       If True, the new beliefs are concatenated with the original BeliefsDataFrame.
+                                  If False, the new beliefs are returned (use pd.concat to add them yourself).
+        """
+        if forecaster is None:
+            forecaster = NaiveForecaster(strategy="last")
+        if event_start is not None:
+            if event_time_window != (None, None):
+                raise ValueError(
+                    "Cannot pass both an event start and event start window."
+                )
+            event_time_window = (event_start, event_start + self.event_resolution)
+
+        # Check assumptions
+        if self.sensor.knowledge_time(event_time_window[0]) < belief_time:
+            # No backcasting
+            raise NotImplementedError("Backcasting is not implemented.")
+        if self.lineage.number_of_sources > 1:
+            # Single source only
+            raise NotImplementedError(
+                "Cannot form new beliefs given multi-sourced data. Consider picking 1 source, e.g. using: df = df[df.index.get_level_values('source') == df.lineage.sources[0]]"
+            )
+        if self.lineage.probabilistic_depth != 1:
+            # Deterministic beliefs only
+            raise NotImplementedError(
+                "Cannot form new beliefs given probabilistic data. Consider pre-computing deterministic beliefs, e.g. using: df = df.make_deterministic()"
+            )
+        if self.lineage.number_of_beliefs != self.lineage.number_of_events:
+            # One belief per event
+            raise NotImplementedError(
+                "Cannot form new beliefs given multiple beliefs about the same event. Consider picking a single belief per event, e.g. using: df = belief_utils.select_most_recent_belief(df)"
+            )
+
+        df = self
+
+        # Do NOT use future data to predict NOR to fit
+        belief_horizon_in_index = False
+        if "belief_horizon" in df.index.names:
+            belief_horizon_in_index = True
+            df = df.convert_index_from_belief_horizon_to_time()
+        df = df[df.index.get_level_values("belief_time") <= belief_time]
+
+        if df.empty:
+            # skip sktime, which expects a non-empty frame
+            y_pred = pd.Series(
+                data=np.nan,
+                index=belief_utils.initialize_index(
+                    start=event_time_window[0],
+                    end=event_time_window[1],
+                    resolution=df.event_resolution,
+                ),
+            )
+        else:
+            # Simplify index
+            df = df.reset_index()[["event_start", "event_value"]].set_index(
+                "event_start"
+            )
+
+            # Convert to local time in UTC, which sktime expects
+            tz = event_time_window[0].tzinfo
+            utc = pytz.utc
+            event_time_window = tuple(
+                dt.astimezone(utc).replace(tzinfo=None) for dt in event_time_window
+            )
+            df.index = df.index.tz_convert(utc).tz_localize(None)
+
+            # Resample to the resolution the data already has, just to set the index frequency, which sktime expects
+            # The new index ends just before the start of the first forecast (which may introduce trailing NaN values),
+            # as sktime expects no gap between the indices of the input and forecasts when applying seasonal periodicity.
+            df = df.reindex(
+                belief_utils.initialize_index(
+                    df.index[0],
+                    event_time_window[0],
+                    resolution=df.event_resolution,
+                )
+            )
+            df = df.resample(df.event_resolution).mean()
+
+            # todo: if forecaster does not handle missing values, impute intermediate missing values and forecast trailing missing values
+
+            # Apply model
+            if isinstance(forecaster, BaseForecaster):
+                forecast_event_starts = belief_utils.initialize_index(
+                    event_time_window[0],
+                    event_time_window[1],
+                    resolution=df.event_resolution,
+                )
+                forecaster.fit(
+                    df["event_value"], df.loc[:, df.columns != "event_value"]
+                )
+                y_pred = forecaster.predict(forecast_event_starts)
+            else:
+                raise NotImplementedError("Consider opening a GitHub issue.")
+
+            # Relocalize to requested timezone
+            y_pred.index = y_pred.index.tz_localize("utc").tz_convert(tz)
+
+        # Prepare results as BeliefsDataframe
+        new_df = BeliefsDataFrame(
+            y_pred, source=source, belief_time=belief_time, sensor=df.sensor
+        )
+        if belief_horizon_in_index:
+            new_df = new_df.convert_index_from_belief_time_to_horizon()
+        if concatenate:
+            return pd.concat([self, new_df])
+        return new_df
+
     def accuracy(
         self,
         t: Union[datetime, timedelta] = None,
@@ -1637,7 +1846,7 @@ class BeliefsDataFrame(pd.DataFrame):
         intuitive_forecast_horizon: bool = True,
         interpolate: bool = True,
         event_value_range: Tuple[Optional[float], Optional[float]] = (None, None),
-    ) -> alt.LayerChart:
+    ) -> "alt.LayerChart":
         """Visualize the BeliefsDataFrame in an interactive Altair chart.
 
         :param show_accuracy: Set to False to plot time series data only
@@ -1665,6 +1874,8 @@ class BeliefsDataFrame(pd.DataFrame):
         Altair: Declarative Visualization in Python.
             https://altair-viz.github.io
         """
+        from timely_beliefs.visualization import utils as visualization_utils
+
         return visualization_utils.plot(
             self,
             show_accuracy=show_accuracy,
@@ -1682,7 +1893,7 @@ class BeliefsDataFrame(pd.DataFrame):
         future_only: bool = False,
         distribution: str = "uniform",
         event_value_window: Tuple[float, float] = None,
-    ) -> alt.FacetChart:
+    ) -> "alt.FacetChart":
         """Create a ridgeline plot of the latest beliefs held at a certain reference time.
 
         :param reference_time: datetime, reference to determine belief horizons
@@ -1692,6 +1903,8 @@ class BeliefsDataFrame(pd.DataFrame):
         :param event_value_window: optional tuple specifying an event value window for the x-axis
                (e.g. plot temperatures between -1 and 21 degrees Celsius)
         """
+        from timely_beliefs.visualization import utils as visualization_utils
+
         if df.lineage.number_of_sources > 1:
             raise ValueError(
                 "Cannot create plot beliefs from multiple sources. BeliefsDataFrame must contain beliefs from a single source."
@@ -1714,7 +1927,7 @@ class BeliefsDataFrame(pd.DataFrame):
         past_only: bool = False,
         distribution: str = "uniform",
         event_value_window: Tuple[float, float] = None,
-    ) -> alt.FacetChart:
+    ) -> "alt.FacetChart":
         """Create a ridgeline plot of the belief history about a specific event.
 
         :param event_start: datetime, indicating the start time of the event for which to plot the belief history
@@ -1724,6 +1937,8 @@ class BeliefsDataFrame(pd.DataFrame):
         :param event_value_window: optional tuple specifying an event value window for the x-axis
                (e.g. plot temperatures between -1 and 21 degrees Celsius)
         """
+        from timely_beliefs.visualization import utils as visualization_utils
+
         if df.lineage.number_of_sources > 1:
             raise ValueError(
                 "Cannot create plot beliefs from multiple sources. BeliefsDataFrame must contain beliefs from a single source."
