@@ -4,6 +4,7 @@ from typing import List, Optional, Union
 
 import numpy as np
 import pandas as pd
+from packaging import version
 from pandas.core.groupby import DataFrameGroupBy
 
 from timely_beliefs import BeliefSource, Sensor
@@ -133,12 +134,10 @@ def respect_event_resolution(grouper: DataFrameGroupBy, resolution):
         # Get the BeliefsDataFrame for a unique belief time and source
         df_slice = group[1]
         if not df_slice.empty:
-            lvl0 = pd.date_range(
+            lvl0 = initialize_index(
                 start=bin_start,
                 end=bin_end,
-                freq=resolution,
-                closed="left",
-                name="event_start",
+                resolution=resolution,
             )
             df = df.append(
                 tb_utils.replace_multi_index_level(
@@ -149,13 +148,46 @@ def respect_event_resolution(grouper: DataFrameGroupBy, resolution):
     return df
 
 
+def propagate_beliefs(
+    df: "classes.BeliefsDataFrame",
+) -> "classes.BeliefsDataFrame":
+    """Propagate beliefs over time by filling NaN values.
+
+    We do this by assuming beliefs propagate over time (ceteris paribus, you still believe what you believed before).
+    That is, the most recent belief about an event is valid until a new belief is formed.
+    If no previous belief has been formed for a certain event, the original NaN valued row will be kept.
+    Requires deterministic data.
+
+    For example:
+
+                                                                                          temp_air  wind_speed  cloud_cover
+    event_start               belief_time               source    cumulative_probability
+    2022-07-01 02:00:00+02:00 2022-06-30 15:10:30+02:00 simulator 0.5                          NaN         NaN         0.94
+                              2022-07-01 01:10:41+02:00 simulator 0.5                        13.61         NaN          NaN
+                              2022-07-01 02:10:32+02:00 simulator 0.5                          NaN        2.88          NaN
+
+    Becomes:
+
+                                                                                          temp_air  wind_speed  cloud_cover
+    event_start               belief_time               source    cumulative_probability
+    2022-07-01 02:00:00+02:00 2022-06-30 15:10:30+02:00 simulator 0.5                          NaN         NaN         0.94
+                              2022-07-01 01:10:41+02:00 simulator 0.5                        13.61         NaN         0.94
+                              2022-07-01 02:10:32+02:00 simulator 0.5                        13.61        2.88         0.94
+    """
+    if df.lineage.probabilistic_depth != 1:
+        raise NotImplementedError(
+            "Propagating probabilistic beliefs is not yet implemented. Please file a GitHub issue."
+        )
+    return df.groupby(level=["event_start", "source"]).ffill()
+
+
 def align_belief_times(
     slice: "classes.BeliefsDataFrame", unique_belief_times
 ) -> "classes.BeliefsDataFrame":
-    """Align belief times such that each event has the same set of unique belief times. We do this by assuming beliefs
-    propagate over time (ceteris paribus, you still believe what you believed before).
-    The most recent belief about an event is valid until a new belief is formed.
-    If no previous belief has been formed, a row is still explicitly included with a nan value.
+    """Align belief times such that each event has the same set of unique belief times.
+    We do this by assuming beliefs propagate over time (ceteris paribus, you still believe what you believed before).
+    That is, the most recent belief about an event is valid until a new belief is formed.
+    If no previous belief has been formed, a row is still explicitly included with a NaN value.
     The input BeliefsDataFrame should represent beliefs about a single event formed by a single source.
     """
 
@@ -492,6 +524,9 @@ def read_csv(
     belief_horizon: timedelta = None,
     belief_time: datetime = None,
     cumulative_probability: float = None,
+    resample: bool = False,
+    timezone: Optional[str] = None,
+    filter_by_column: dict = None,
     **kwargs,
 ) -> "classes.BeliefsDataFrame":
     """Utility function to load a BeliefsDataFrame from a csv file or xls sheet (see example/temperature.csv).
@@ -500,11 +535,38 @@ def read_csv(
     In case the csv file contains multiple source names, you can pass a list of sources.
     Each source name will be replaced by the actual source.
 
+    :param path:                    Path (or url) to the csv file.
+    :param sensor:                  The sensor to which the data pertains.
+    :param source:                  Optionally, set a specific source for the read-in data.
+                                    If not set, the look_up_sources parameter must be used.
+    :param look_up_sources:         Optionally, pass a list of sources used to look up source names from the csv file.
+                                    If not set, the source parameter must be used.
+    :param belief_horizon:          Optionally, set a specific belief horizon for the read-in data.
+    :param belief_time:             Optionally, set a specific belief time for the read-in data.
+    :param cumulative_probability:  Optionally, set a specific cumulative probability for the read-in data.
+    :param resample:                Optionally, resample to the event resolution of the sensor.
+                                    Only implemented for the special read case of 2-column data (see below).
+    :param timezone:                Optionally, localize timezone naive datetimes to a specific timezone.
+                                    Accepts IANA timezone names (e.g. UTC or Europe/Amsterdam).
+                                    If not set and timezone naive datetimes are read in, the data is localized to UTC.
+    :param filter_by_column:        Select a subset of rows by filtering on a specific value for a specific column.
+                                    For example: {4: 1995} selects all rows where column 4 contains the value 1995.
+
     Also supports the case of a csv file with just 2 columns and 1 header row (a quite common time series format).
-    In this case no special header names are required, but the first column has to contain the UTC event starts,
+    In this case no special header names are required, but the first column has to contain the event starts,
     and the second column has to contain the event values.
+    In case the event starts are local times (timezone naive), they must be localized to a timezone,
+    by passing the relevant IANA timezone name (e.g. timezone='UTC' or timezone='Europe/Amsterdam').
     You also need to pass explicit values for the belief horizon/time and cumulative probability,
     in addition to the sensor and source.
+    If needed, the time series may be resampled to the event resolution of the sensor, using resample=True.
+    This is the only case that supports resampling.
+
+    Also supports the case of a csv file with just 3 columns and 1 header row.
+    In this case no special header names are required, but the first two columns have to contain the event starts
+    and belief times, respectively, and the third column has to contain the event values.
+    In case event starts and belief times are local times (timezone naive), they must be localized to a timezone,
+    by passing the relevant IANA timezone name (e.g. timezone='UTC' or timezone='Europe/Amsterdam').
 
     Consult pandas documentation for which additional kwargs can be passed to pandas.read_csv or pandas.read_excel.
     Useful examples are parse_dates=True, infer_datetime_format=True (for read_csv)
@@ -515,6 +577,14 @@ def read_csv(
     >>> df.to_csv()
 
     """
+    original_usecols = kwargs.get("usecols", []).copy()
+    if filter_by_column:
+        # Also read in any extra columns used to filter the read-in data
+        kwargs["usecols"] += [
+            col
+            for col in filter_by_column.keys()
+            if col not in kwargs.get("usecols", [])
+        ]
     ext = path.split(".")[-1]
     if ext.lower() == "csv":
         df = pd.read_csv(path, **kwargs)
@@ -524,14 +594,28 @@ def read_csv(
         raise TypeError(
             f"Extension {ext} not recognized. Accepted file extensions are csv, xlsx and xls."
         )
-
-    # Special case for simple time series (UTC datetime in 1st column and value in 2nd column)
-    if len(df.columns) == 2:
-        df.columns = ["event_start", "event_value"]
-        df["event_start"] = pd.to_datetime(df["event_start"], utc=True).dt.tz_convert(
-            sensor.timezone
+    if filter_by_column:
+        # Filter the read-in data
+        for col, val in filter_by_column.items():
+            df = df[df[col] == val]
+        # Remove the extra columns used to filter
+        df = df.drop(
+            columns=[
+                col for col in filter_by_column.keys() if col not in original_usecols
+            ]
         )
 
+    # Exclude rows with NaN or NaT values
+    if not kwargs.get("keep_default_na", True):
+        df = df.dropna()
+
+    # Preserve order of usecols
+    if "usecols" in kwargs:
+        df = df[[col for col in kwargs["usecols"] if col in df.columns]]
+
+    # Special cases for simple time series
+    df = interpret_special_read_cases(df, sensor, resample, timezone)
+
     # Apply optionally set belief timing
     if belief_horizon is not None and belief_time is not None:
         raise ValueError("Cannot set both a belief horizon and a belief time.")
@@ -540,7 +624,30 @@ def read_csv(
     elif belief_time is not None:
         df["belief_time"] = belief_time
 
-    # Apply optionally set source
+    # Apply optionally set source, or look up sources
+    df = fill_in_sources(df, source, look_up_sources, ext)
+
+    # Apply optionally set cumulative probability
+    if cumulative_probability is not None:
+        df["cumulative_probability"] = cumulative_probability
+
+    return classes.BeliefsDataFrame(df, sensor=sensor)
+
+
+def fill_in_sources(
+    df: pd.DataFrame,
+    source: Union["classes.BeliefSource", None],
+    look_up_sources: Union[List["classes.BeliefSource"], None],
+    ext: str,
+) -> pd.DataFrame:
+    """Fill the 'source' column with BeliefSource objects.
+
+    :param df:              DataFrame whose 'source' column (if there is one), only contains source names.
+    :param source:          If set, the 'source' column is filled with this BeliefSource.
+    :param look_up_sources: If set, the source names in the 'source' column are replaced
+                            with the corresponding BeliefSource from this list.
+    :param ext:             File extension as a string, used in warning or error messages.
+    """
     if source is not None:
         df["source"] = source_utils.ensure_source_exists(source)
     elif "source" in df.columns:
@@ -555,13 +662,100 @@ def read_csv(
                 f"Sources are stored in {ext} file by their name or id. Please specify a list of BeliefSources for looking them up."
             )
     else:
-        raise Exception(f"No source specified in {ext}, please set a source.")
+        raise Exception(f"No source specified in {ext} file, please set a source.")
+    return df
 
-    # Apply optionally set cumulative probability
-    if cumulative_probability is not None:
-        df["cumulative_probability"] = cumulative_probability
 
-    return classes.BeliefsDataFrame(df, sensor=sensor)
+def interpret_special_read_cases(
+    df: pd.DataFrame, sensor: "classes.Sensor", resample: bool, timezone: Optional[str]
+) -> pd.DataFrame:
+    """Interpret the read-in data, either as event starts and event values (2 cols),
+    or as event starts, belief times and event values (3 cols).
+    """
+    if resample and len(df.columns) != 2:
+        raise NotImplementedError("Resampling is not supported for this import case.")
+
+    if len(df.columns) == 2:
+        # datetime in 1st column and value in 2nd column
+        df.columns = ["event_start", "event_value"]
+        df["event_start"] = convert_to_timezone(
+            df["event_start"],
+            timezone_to_convert_to=sensor.timezone,
+            timezone_to_localize_to=timezone,
+        )
+        if resample:
+            df = (
+                df.set_index("event_start")
+                .resample(sensor.event_resolution)
+                .mean()
+                .reset_index()
+            )
+    elif len(df.columns) == 3:
+        # datetimes in 1st and 2nd column, and value in 3rd column
+        df.columns = ["event_start", "belief_time", "event_value"]
+        df["event_start"] = convert_to_timezone(
+            df["event_start"],
+            timezone_to_convert_to=sensor.timezone,
+            timezone_to_localize_to=timezone,
+        )
+        df["belief_time"] = convert_to_timezone(
+            df["belief_time"],
+            timezone_to_convert_to=sensor.timezone,
+            timezone_to_localize_to=timezone,
+        )
+    return df
+
+
+def convert_to_timezone(
+    s: pd.Series, timezone_to_convert_to: str, timezone_to_localize_to: Optional[str]
+) -> pd.Series:
+    """Convert the timezone of the series to the given timezone.
+
+    In case the series contains naive datetimes, they are first localized using the 'timezone_to_localize_to' parameter.
+
+    :param s:                           series with datetime representations
+    :param timezone_to_convert_to:      timezone to which all datetimes are converted
+    :param timezone_to_localize_to:     optional timezone for localizing timezone naive datetimes
+    :raises:                            TypeError in case naive datetimes are passed without a timezone to localize to
+    """
+    # Convert to datetime (works for timezone naive datetimes, and timezone aware datetime with a shared offset)
+    s = pd.to_datetime(s)
+    if s.dtype == "object":
+        # Reattempt conversion for timezone aware datetimes with a mixed offset
+        s = pd.to_datetime(s, utc=True)
+    if s.dt.tz is None:
+        if timezone_to_localize_to is None:
+            raise TypeError(
+                f"The timely-beliefs package does not work with timezone-naive datetimes. Please specify a timezone to which to localize your data (e.g. the timezone of the sensor, which is '{timezone_to_convert_to}')."
+            )
+        elif timezone_to_localize_to != timezone_to_convert_to:
+            warnings.warn(
+                f"Converting the timezone of the data from {timezone_to_localize_to} to {timezone_to_convert_to}."
+            )
+        s = s.dt.tz_localize(timezone_to_localize_to, ambiguous="infer")
+    return s.dt.tz_convert(timezone_to_convert_to)
+
+
+def initialize_index(
+    start: datetime, end: datetime, resolution: timedelta, inclusive: str = "left"
+) -> pd.DatetimeIndex:
+    """Initialize DatetimeIndex for event starts.
+
+    Supports updated function signature of pd.date_range.
+    From pandas>=1.4.0, it is clear that 'closed' will be replaced by 'inclusive'.
+    """
+    if version.parse(pd.__version__) >= version.parse("1.4.0"):
+        return pd.date_range(
+            start=start,
+            end=end,
+            freq=resolution,
+            inclusive=inclusive,
+            name="event_start",
+        )
+    else:
+        return pd.date_range(
+            start=start, end=end, freq=resolution, closed=inclusive, name="event_start"
+        )
 
 
 def is_pandas_structure(x):
@@ -570,3 +764,15 @@ def is_pandas_structure(x):
 
 def is_tb_structure(x):
     return isinstance(x, (classes.BeliefsDataFrame, classes.BeliefsSeries))
+
+
+def extreme_timedeltas_not_equal(
+    td_a: Union[timedelta, pd.Timedelta],
+    td_b: timedelta,
+) -> bool:
+    """Workaround for pd.Timedelta(...) != timedelta.max (or min)
+    See pandas GH49021.
+    """
+    if isinstance(td_a, pd.Timedelta):
+        td_a = td_a.to_pytimedelta()
+    return td_a != td_b
