@@ -7,13 +7,17 @@ from pytz import utc
 from timely_beliefs import DBBeliefSource, DBSensor, DBTimedBelief
 from timely_beliefs.db_base import Base
 from timely_beliefs.sensors.func_store.knowledge_horizons import (
+    at_date,
     determine_ex_ante_knowledge_horizon_for_x_days_ago_at_y_oclock,
 )
 from timely_beliefs.tests import engine, session
 
 if sys.version_info[0] == 3 and sys.version_info[1] == 6:
     # Ignore these tests for python==3.6
-    collect_ignore = ["test_ignore_36.py"]
+    collect_ignore = ["test_ignore_36.py", "test_plotting.py", "test_ridgeline.py"]
+if sys.version_info[0] == 3 and sys.version_info[1] == 7:
+    # Ignore these tests for python==3.7
+    collect_ignore = ["test_plotting.py", "test_ridgeline.py"]
 
 
 @pytest.fixture(scope="function")
@@ -80,6 +84,22 @@ def create_ex_post_time_slot_sensor(name: str) -> DBSensor:
     return sensor
 
 
+@pytest.fixture(scope="function", autouse=False)
+def unique_knowledge_time_sensor() -> DBSensor:
+    """Define sensor recording events with a unique knowledge time."""
+    sensor = DBSensor(
+        name="SinglePublicationSensor",
+        event_resolution=timedelta(hours=1),
+        knowledge_horizon=(
+            at_date,
+            dict(knowledge_time=datetime(1990, 5, 10, 0, tzinfo=utc)),
+        ),
+    )
+    session.add(sensor)
+    session.flush()
+    return sensor
+
+
 @pytest.fixture(scope="function", autouse=True)
 def test_source_a(db):
     """Define source for test beliefs."""
