@@ -1,7 +1,12 @@
 from datetime import timedelta
 from typing import List, Optional, Tuple
 
-import altair as alt
+try:
+    import altair as alt
+except ModuleNotFoundError:
+    raise ModuleNotFoundError(
+        "Altair is a soft dependency and not included in the base timely-beliefs installation. Please run `pip install altair` to install the altair package. To install all soft dependencies for visualization, run: `pip install timely_beliefs[viz]`."
+    )
 import numpy as np
 import pandas as pd
 
@@ -158,7 +163,7 @@ def plot(
 
 def timedelta_to_human_range(s: pd.Series) -> Tuple[pd.Series, str]:
     """Convert a pandas Series of timedeltas to a pandas Series of floats,
-    and derive a time unit (a string such as "years" or "minutes") that gives a nice human readable range of floats.
+    and derive a time unit (a string such as "years" or "minutes") that gives a nice human-readable range of floats.
     For example:
 
     >>> import timely_beliefs as tb
@@ -177,7 +182,7 @@ def timedelta_to_human_range(s: pd.Series) -> Tuple[pd.Series, str]:
         217   3 days 01:00:00
         Length: 74, dtype: timedelta64[ns]
     >>> from timely_beliefs.visualization.utils import timedelta_to_human_range
-    >>> timedelta_to_human_range(horizons)  # This is human readable range, though
+    >>> timedelta_to_human_range(horizons)  # This is a human-readable range, though
     <<< (0       0.0
         1       1.0
         4       2.0
@@ -191,35 +196,37 @@ def timedelta_to_human_range(s: pd.Series) -> Tuple[pd.Series, str]:
         217    73.0
         Length: 74, dtype: float64, 'hours')
     """
+    if s.empty:
+        return s, "hours"
     timedelta_span = max(s) - min(s)
     if timedelta_span >= timedelta(days=4 * 365.2425):
         s = s.apply(
             lambda x: x.days / 365.2425
-            + (x.seconds + x.microseconds / 10 ** 6) / (365.2425 * 24 * 60 * 60)
+            + (x.seconds + x.microseconds / 10**6) / (365.2425 * 24 * 60 * 60)
         )
         time_unit = "years"
     elif timedelta_span >= timedelta(days=4):
         s = s.apply(
-            lambda x: x.days + (x.seconds + x.microseconds / 10 ** 6) / (24 * 60 * 60)
+            lambda x: x.days + (x.seconds + x.microseconds / 10**6) / (24 * 60 * 60)
         )
         time_unit = "days"
     elif timedelta_span >= timedelta(hours=4):
         s = s.apply(
-            lambda x: x.days * 24 + (x.seconds + x.microseconds / 10 ** 6) / (60 * 60)
+            lambda x: x.days * 24 + (x.seconds + x.microseconds / 10**6) / (60 * 60)
         )
         time_unit = "hours"
     elif timedelta_span >= timedelta(minutes=4):
         s = s.apply(
-            lambda x: x.days * 24 * 60 + (x.seconds + x.microseconds / 10 ** 6) / 60
+            lambda x: x.days * 24 * 60 + (x.seconds + x.microseconds / 10**6) / 60
         )
         time_unit = "minutes"
     elif timedelta_span >= timedelta(seconds=4):
         s = s.apply(
-            lambda x: x.days * 24 * 60 * 60 + x.seconds + x.microseconds / 10 ** 6
+            lambda x: x.days * 24 * 60 * 60 + x.seconds + x.microseconds / 10**6
         )
         time_unit = "seconds"
     else:
-        s = s.apply(lambda x: x.days * 24 * 60 * 60 * 10 ** 6 + x.microseconds)
+        s = s.apply(lambda x: x.days * 24 * 60 * 60 * 10**6 + x.microseconds)
         time_unit = "microseconds"
     return s, time_unit
 
